/* stereo_calib.cpp
 * Modified version that read couples of images from two
 * cameras connected to the computer, they are the video devices '0 for left
 * image and '1' for right image */

/* Giandomenico De Sanctis @ 2/2012
 * gidesay@yahoo.com */

/*----------------------------------------------------------------------------- */
/* The original code was from the OpenCV book. The copyright notice is below */

/* *************** License:**************************
   Oct. 3, 2008
   Right to use this code in any way you want without warranty, support or any guarantee of it working.

   BOOK: It would be nice if you cited it:
   Learning OpenCV: Computer Vision with the OpenCV Library
     by Gary Bradski and Adrian Kaehler
     Published by O'Reilly Media, October 3, 2008
 
   AVAILABLE AT: 
     http://www.amazon.com/Learning-OpenCV-Computer-Vision-Library/dp/0596516134
     Or: http://oreilly.com/catalog/9780596516130/
     ISBN-10: 0596516134 or: ISBN-13: 978-0596516130    

   OTHER OPENCV SITES:
   * The source code is on sourceforge at:
     http://sourceforge.net/projects/opencvlibrary/
   * The OpenCV wiki page (As of Oct 1, 2008 this is down for changing over servers, but should come back):
     http://opencvlibrary.sourceforge.net/
   * An active user group is at:
     http://tech.groups.yahoo.com/group/OpenCV/
   * The minutes of weekly OpenCV development meetings are at:
     http://pr.willowgarage.com/wiki/OpenCV
   ************************************************** */

#include "opencv2/calib3d/calib3d.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/imgproc/imgproc_c.h"

#include <vector>
#include <string>
#include <algorithm>
#include <iostream>
#include <iterator>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

using namespace cv;
using namespace std;

int print_help()
{
	cout <<
			" Directly generate a series of chessboard images, from left camera (video device=0) \n"
			" and right camera (video device=1).\n"
			" With that chessboard images, and the number of corners (w, h)\n"
			" on the chessboards, \n"
			" calibrate the cameras, save the computed coefficients matrix, and display the\n"
			" rectified results.	\n" << endl;
    cout << "Usage:\n stereo_calib -w <board_width> -h <board_height> [-nc <num board images>] [-useuncalibrated /* use Hartley rectification method */]  [-nr /*do not view results*/] \n" << endl;
    return 0;
}


static void
StereoCalib(const vector<string>& imagelist, Size boardSize, bool useCalibrated=true, bool showRectified=true)
{
    if( imagelist.size() % 2 != 0 )
    {
        cout << "Error: the image list contains odd (non-even) number of elements\n";
        return;
    }
    
    bool displayCorners = false;
    const int maxScale = 2;
    const float squareSize = 1.f;  // Set this to your actual square size
    // ARRAY AND VECTOR STORAGE:
    
    vector<vector<Point2f> > imagePoints[2];
    vector<vector<Point3f> > objectPoints;
    Size imageSize;
    
    int i, j, k, nimages = (int)imagelist.size()/2;
    
    imagePoints[0].resize(nimages);
    imagePoints[1].resize(nimages);
    vector<string> goodImageList;
    
    for( i = j = 0; i < nimages; i++ )
    {
        for( k = 0; k < 2; k++ )
        {
            const string& filename = imagelist[i*2+k];
            Mat img = imread(filename, 0);
            if(img.empty())
                break;
            if( imageSize == Size() )
                imageSize = img.size();
            else if( img.size() != imageSize )
            {
                cout << "The image " << filename << " has the size different from the first image size. Skipping the pair\n";
                break;
            }
            bool found = false;
            vector<Point2f>& corners = imagePoints[k][j];
            for( int scale = 1; scale <= maxScale; scale++ )
            {
                Mat timg;
                if( scale == 1 )
                    timg = img;
                else
                    resize(img, timg, Size(), scale, scale);
                found = findChessboardCorners(timg, boardSize, corners, 
                    CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_NORMALIZE_IMAGE);
                if( found )
                {
                    if( scale > 1 )
                    {
                        Mat cornersMat(corners);
                        cornersMat *= 1./scale;
                    }
                    break;
                }
            }
            if( displayCorners )
            {
                cout << filename << endl;
                Mat cimg, cimg1;
                cvtColor(img, cimg, CV_GRAY2BGR);
                drawChessboardCorners(cimg, boardSize, corners, found);
                double sf = 640./MAX(img.rows, img.cols);
                resize(cimg, cimg1, Size(), sf, sf);
                imshow("corners", cimg1);
                char c = (char)waitKey(500);
                if( c == 27 || c == 'q' || c == 'Q' ) //Allow ESC to quit
                    exit(-1);
            }
            else
                putchar('.');
            if( !found )
                break;
            cornerSubPix(img, corners, Size(11,11), Size(-1,-1),
                         TermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS,
                                      30, 0.01));
        }
        if( k == 2 )
        {
            goodImageList.push_back(imagelist[i*2]);
            goodImageList.push_back(imagelist[i*2+1]);
            j++;
        }
    }
    cout << j << " pairs have been successfully detected.\n";
    nimages = j;
    if( nimages < 2 )
    {
        cout << "Error: too little pairs to run the calibration\n";
        return;
    }
    
    imagePoints[0].resize(nimages);
    imagePoints[1].resize(nimages);
    objectPoints.resize(nimages);
    
    for( i = 0; i < nimages; i++ )
    {
        for( j = 0; j < boardSize.height; j++ )
            for( k = 0; k < boardSize.width; k++ )
                objectPoints[i].push_back(Point3f(j*squareSize, k*squareSize, 0));
    }
    
    cout << "Running stereo calibration ...\n";
    
    Mat cameraMatrix[2], distCoeffs[2];
    cameraMatrix[0] = Mat::eye(3, 3, CV_64F);
    cameraMatrix[1] = Mat::eye(3, 3, CV_64F);
    Mat R, T, E, F;
    int flags =     CV_CALIB_FIX_ASPECT_RATIO +
                    CV_CALIB_ZERO_TANGENT_DIST +
                    CV_CALIB_RATIONAL_MODEL +
                    CV_CALIB_FIX_K3 + CV_CALIB_FIX_K4 + CV_CALIB_FIX_K5;
;
	// the calibration assuming same focal length for the two cameras
	// doesn't work if we have real different focal length and use the
	// calibrated rectification method (Bouguet method)
	if (useCalibrated == false)
	{
		flags = flags + CV_CALIB_SAME_FOCAL_LENGTH;
	};

    double rms = stereoCalibrate(objectPoints, imagePoints[0], imagePoints[1],
                    cameraMatrix[0], distCoeffs[0],
                    cameraMatrix[1], distCoeffs[1],
                    imageSize, R, T, E, F,
                    TermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS, 100, 1e-5),
					flags) ;
    cout << "done with RMS error=" << rms << endl;



// CALIBRATION QUALITY CHECK
// because the output fundamental matrix implicitly
// includes all the output information,
// we can check the quality of calibration using the
// epipolar geometry constraint: m2^t*F*m1=0
    double err = 0;
    int npoints = 0;
    vector<Vec3f> lines[2];
    for( i = 0; i < nimages; i++ )
    {
        int npt = (int)imagePoints[0][i].size();
        Mat imgpt[2];
        for( k = 0; k < 2; k++ )
        {
            imgpt[k] = Mat(imagePoints[k][i]);
            undistortPoints(imgpt[k], imgpt[k], cameraMatrix[k], distCoeffs[k], Mat(), cameraMatrix[k]);
            computeCorrespondEpilines(imgpt[k], k+1, F, lines[k]);
        }
        for( j = 0; j < npt; j++ )
        {
            double errij = fabs(imagePoints[0][i][j].x*lines[1][j][0] +
                                imagePoints[0][i][j].y*lines[1][j][1] + lines[1][j][2]) +
                           fabs(imagePoints[1][i][j].x*lines[0][j][0] +
                                imagePoints[1][i][j].y*lines[0][j][1] + lines[0][j][2]);
            err += errij;
        }
        npoints += npt;
    }
    cout << "average reprojection err = " <<  err/npoints << endl;
    
    // save intrinsic parameters
    FileStorage fs("intrinsics.xml", CV_STORAGE_WRITE);
    if( fs.isOpened() )
    {
        fs << "M1" << cameraMatrix[0] << "D1" << distCoeffs[0] <<
            "M2" << cameraMatrix[1] << "D2" << distCoeffs[1];
        fs.release();
    }
    else
        cout << "Error: can not save the intrinsic parameters\n";
    
    Mat R1, R2, P1, P2, Q;
    Rect validRoi[2];
    Mat rmap[2][2];
	bool isVerticalStereo;

// IF BY CALIBRATED (BOUGUET'S METHOD)
    if( useCalibrated )
    {
		cv::Size newImageSize = imageSize;
		stereoRectify(cameraMatrix[0], distCoeffs[0],
			          cameraMatrix[1], distCoeffs[1],
				      imageSize, R, T, R1, R2, P1, P2, Q,
					  CALIB_ZERO_DISPARITY, 1, newImageSize, &validRoi[0], &validRoi[1]);

    // OpenCV can handle left-right
    // or up-down camera arrangements
		isVerticalStereo = fabs(P2.at<double>(1, 3)) > fabs(P2.at<double>(0, 3));
   }
// OR ELSE HARTLEY'S METHOD
    else
 // use intrinsic parameters of each camera, but
 // compute the rectification transformation directly
 // from the fundamental matrix
    {
        vector<Point2f> allimgpt[2];
        for( k = 0; k < 2; k++ )
        {
            for( i = 0; i < nimages; i++ )
                std::copy(imagePoints[k][i].begin(), imagePoints[k][i].end(), back_inserter(allimgpt[k]));
        }
        F = findFundamentalMat(Mat(allimgpt[0]), Mat(allimgpt[1]), FM_8POINT, 0, 0);
		
		Mat H1, H2;
        stereoRectifyUncalibrated(Mat(allimgpt[0]), Mat(allimgpt[1]), F, imageSize, H1, H2, 3);
        
        R1 = cameraMatrix[0].inv()*H1*cameraMatrix[0];
        R2 = cameraMatrix[1].inv()*H2*cameraMatrix[1];
        P1 = cameraMatrix[0];
        P2 = cameraMatrix[1];
		Q  = Mat(4, 4, CV_64F);
		Q.zeros(4,4,CV_64F);


		isVerticalStereo = false;    
	}

    if( useCalibrated )
	{    fs.open("extrinsics.xml", CV_STORAGE_WRITE);
	}
	else
	{    fs.open("extrinsics-uncalib.xml", CV_STORAGE_WRITE);
	}

    if( fs.isOpened() )
    {
        fs << "R" << R << "T" << T << "R1" << R1 << "R2" << R2 << "P1" << P1 << "P2" << P2 << "Q" << Q;
        fs.release();
    }
    else
        cout << "Error: can not save the intrinsic parameters\n";

	// COMPUTE AND DISPLAY RECTIFICATION
    if( !showRectified )
        return;

    //Precompute maps for cv::remap()
    initUndistortRectifyMap(cameraMatrix[0], distCoeffs[0], R1, P1, imageSize, CV_16SC2, rmap[0][0], rmap[0][1]);
    initUndistortRectifyMap(cameraMatrix[1], distCoeffs[1], R2, P2, imageSize, CV_16SC2, rmap[1][0], rmap[1][1]);
    
    Mat canvas;
    double sf;
    int w, h;
    if( !isVerticalStereo )
    {
        sf = 600./MAX(imageSize.width, imageSize.height);
        w = cvRound(imageSize.width*sf);
        h = cvRound(imageSize.height*sf);
        canvas.create(h, w*2, CV_8UC3);
    }
    else
    {
        sf = 300./MAX(imageSize.width, imageSize.height);
        w = cvRound(imageSize.width*sf);
        h = cvRound(imageSize.height*sf);
        canvas.create(h*2, w, CV_8UC3);
    }
    
    for( i = 0; i < nimages; i++ )
    {
        for( k = 0; k < 2; k++ )
        {
            Mat img = imread(goodImageList[i*2+k], 0), rimg, cimg;
            remap(img, rimg, rmap[k][0], rmap[k][1], CV_INTER_LINEAR);
            cvtColor(rimg, cimg, CV_GRAY2BGR);
            Mat canvasPart = !isVerticalStereo ? canvas(Rect(w*k, 0, w, h)) : canvas(Rect(0, h*k, w, h));
            resize(cimg, canvasPart, canvasPart.size(), 0, 0, CV_INTER_AREA);
            if( useCalibrated )
            {
                Rect vroi(cvRound(validRoi[k].x*sf), cvRound(validRoi[k].y*sf),
                          cvRound(validRoi[k].width*sf), cvRound(validRoi[k].height*sf)); 
                rectangle(canvasPart, vroi, Scalar(0,0,255), 3, 8);
            }
        }
        
        if( !isVerticalStereo )
            for( j = 0; j < canvas.rows; j += 16 )
                line(canvas, Point(0, j), Point(canvas.cols, j), Scalar(0, 255, 0), 1, 8);
        else
            for( j = 0; j < canvas.cols; j += 16 )
                line(canvas, Point(j, 0), Point(j, canvas.rows), Scalar(0, 255, 0), 1, 8);
        imshow("rectified", canvas);
        char c = (char)waitKey();
        if( c == 27 || c == 'q' || c == 'Q' )
            break;
    }
}

                   

static void findChessboard(const char winName[], const IplImage* image, const IplImage* gray_image,
	       	const CvSize board_sz, int* count, CvPoint2D32f* corners)
{
	// Find chessboard corners:
	int found = cvFindChessboardCorners( image,   board_sz, corners,
		count , CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_NORMALIZE_IMAGE );
	// Get subpixel accuracy on those corners
	cvCvtColor( image, (CvArr*) gray_image, CV_BGR2GRAY );
	cvFindCornerSubPix( gray_image, corners, *count, cvSize( 11, 11 ), 
		cvSize( -1, -1 ), cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 30, 0.1 ));

	// Draw chessboard on a temp image, so original image can be saved
	IplImage* temp = cvCloneImage(image);
	cvDrawChessboardCorners( (CvArr*) temp, board_sz, corners, *count, found );
	cvShowImage( winName, temp );
	cvReleaseImage(&temp);
}

static bool readChessImageList( const Size boardsize, const int n_boards, vector<string>& l  )
{
	//int n_boards <- Number of chessboards images required
	const int board_dt = 20;
	bool ret = false;
    
	l.resize(0);

	int board_w = boardsize.width;
	int board_h = boardsize.height;
	int board_n = board_w * board_h;
	CvSize board_sz = cvSize( board_w, board_h );
	// left
	CvCapture* captureL = cvCaptureFromCAM( 0 );
	assert( captureL );
	cvNamedWindow( "Stereo Calibration -Left" );
	cvMoveWindow( "Stereo Calibration -Left", 5, 5); 

	// right
	CvCapture* captureR = cvCaptureFromCAM( 1 );
	assert( captureR );
	cvNamedWindow( "Stereo Calibration -Right" );
	cvMoveWindow( "Stereo Calibration -Right", 700, 5); 
		

	CvPoint2D32f* corners = new CvPoint2D32f[ board_n ];
	int corner_countL, corner_countR;
	int successes = 0;
	int frame = 0;
	char imageName[80] = ""; 

	// store in buffer the frames from the 2 cameras using the fast function cvGrabFrame,
	// so the frames have the best synchronization
	cvGrabFrame( captureL );
	cvGrabFrame( captureR );


	// retrieve the images from buffer by the (relatively) slow function cvRetrieveFrame	
	IplImage *imageL = cvRetrieveFrame( captureL );
	IplImage *gray_imageL = cvCreateImage( cvGetSize( imageL ), 8, 1 );

	IplImage *imageR = cvRetrieveFrame( captureR );
	IplImage *gray_imageR = cvCreateImage( cvGetSize( imageR ), 8, 1 );

	// Capture Corner views loop until we've got n_boards
	// succesfull captures (all corners on the board are found)

	while( successes < n_boards ){
		// Skip every board_dt frames to allow user to move chessboard
		if( frame++ % board_dt == 0 ){
			// ----left----
 			findChessboard("Stereo Calibration -Left", imageL, gray_imageL,
	       	 				board_sz, &corner_countL,  &corners[0]);

			// ----right----
 			findChessboard("Stereo Calibration -Right", imageR, gray_imageR,
	       	 				board_sz, &corner_countR,  &corners[0]);


			// If we got a good board, add it to our data
			if ( ( corner_countL == board_n ) && ( corner_countR == board_n ) )
			{
				successes++;
				cout << "Successfully acquired " << successes << " chessboard images\n";
				sprintf(imageName , "stereocalib-%c-%2i.JPG", 'L', successes);
				cvSaveImage(imageName, imageL);
			        l.push_back((string)imageName);

				sprintf(imageName , "stereocalib-%c-%2i.JPG", 'R', successes);
				cvSaveImage(imageName, imageR);
			        l.push_back((string)imageName);
			}
		} 

		// Handle pause/unpause and ESC
		int c = cvWaitKey( 15 );
		if( c == 'p' ){
			c = 0;
			while( c != 'p' && c != 27 ){
				c = cvWaitKey( 250 );
			}
		}
		if( c == 27 )
			break;
		
		// Get next couple of images
		cvGrabFrame( captureL );
		cvGrabFrame( captureR );

		imageL = cvRetrieveFrame( captureL );
		imageR = cvRetrieveFrame( captureR );
	} // End collection while loop

	cvReleaseCapture(&captureL);
	cvReleaseCapture(&captureR);
	cvDestroyAllWindows();

	if (successes >= n_boards) {ret = true; }; 
    
    return ret;
}

int main(int argc, char** argv)
{
    Size boardSize;
    string imagelistfn;
    bool showRectified = true;
	bool useCalibrated = true;
	int nBoard = 12; 
    
    for( int i = 1; i < argc; i++ )
    {
        if( string(argv[i]) == "-w" )
        {
            if( sscanf(argv[++i], "%d", &boardSize.width) != 1 || boardSize.width <= 0 )
            {
                cout << "invalid board width" << endl;
                return print_help();
            }
        }
        else if( string(argv[i]) == "-h" )
        {
            if( sscanf(argv[++i], "%d", &boardSize.height) != 1 || boardSize.height <= 0 )
            {
                cout << "invalid board height" << endl;
                return print_help();
            }
        }
        else if( string(argv[i]) == "-nr" )
            showRectified = false;
        else if( string(argv[i]) == "--help" )
            return print_help();
        else if( string(argv[i]) == "-useuncalibrated" )
		{
			useCalibrated = false;
		}
        else if( string(argv[i]) == "-nc" )
        {
            if( sscanf(argv[++i], "%d", &nBoard) != 1 || nBoard < 8 )
            {
                cout << "minimum number of board images must be 8" << endl;
                return print_help();
            }
        }
        else if( argv[i][0] == '-' )
        {
            cout << "invalid option " << argv[i] << endl;
            return 0;
        }
        else
            imagelistfn = argv[i];
    }
    
    if( boardSize.width == 0 )
    {
        boardSize = Size(9, 6);
    }
    else if( boardSize.width <= 0 || boardSize.height <= 0 )
    {
        cout << "you must specify the board width and height (-w and -h options)" << endl; 
        return 0;
    }
    
    vector<string> imagelist;
 
	
    bool ok = readChessImageList(boardSize, nBoard, imagelist);

    if(!ok || imagelist.empty())
    {
        cout << "the string list is empty" << endl;
        return print_help();
    } 
 

	cout << "Executing calibration " ;
    StereoCalib(imagelist, boardSize, useCalibrated, showRectified);
	
	cout << "Press a key" << endl;
	cvWaitKey( 0 );
 
	return 0;
}

